# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The resource filters api."""

from keystoneauth1 import exceptions as ks_exc
from six.moves import http_client
from webob import exc

from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import default_types as default_types
from cinder.api.v3.views import default_types as default_types_view
from cinder.api import validation
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.policies import default_types as policy
from cinder import quota_utils


class DefaultTypesController(wsgi.Controller):
    """The Default types API controller for the OpenStack API."""

    _view_builder_class = default_types_view.ViewBuilder

    def _validate_project_and_authorize(self, context, project_id,
                                        policy_check):
        try:
            target_project = quota_utils.get_project_hierarchy(context,
                                                               project_id)
            target_project = {'project_id': target_project.id,
                              'domain_id': target_project.domain_id}
            context.authorize(policy_check, target=target_project)
        except ks_exc.http.NotFound:
            explanation = _("Project with id %s not found." % project_id)
            raise exc.HTTPNotFound(explanation=explanation)
        except exception.NotAuthorized:
            explanation = _("You are not authorized to perform this "
                            "operation.")
            raise exc.HTTPForbidden(explanation=explanation)

    @wsgi.response(http_client.OK)
    @wsgi.Controller.api_version(mv.DEFAULT_TYPE_OVERRIDES)
    @validation.schema(default_types.create_or_update)
    def create_update(self, req, id, body):
        """Set a default volume type for the specified project."""
        context = req.environ['cinder.context']

        project_id = id
        volume_type_id = body['default_type']['volume_type']

        self._validate_project_and_authorize(context, project_id,
                                             policy.CREATE_UPDATE_POLICY)
        try:
            volume_type_id = objects.VolumeType.get_by_name_or_id(
                context, volume_type_id).id

        except exception.VolumeTypeNotFound as e:
            raise exc.HTTPBadRequest(explanation=e.msg)

        default_type = db.project_default_volume_type_set(
            context, volume_type_id, project_id)

        return self._view_builder.create(default_type)

    @wsgi.response(http_client.OK)
    @wsgi.Controller.api_version(mv.DEFAULT_TYPE_OVERRIDES)
    def detail(self, req, id):
        """Return detail of a default type."""

        context = req.environ['cinder.context']

        project_id = id
        self._validate_project_and_authorize(context, project_id,
                                             policy.GET_POLICY)
        default_type = db.project_default_volume_type_get(context, project_id)
        if not default_type:
            raise exception.VolumeTypeProjectDefaultNotFound(
                project_id=project_id)
        return self._view_builder.detail(default_type)

    @wsgi.response(http_client.OK)
    @wsgi.Controller.api_version(mv.DEFAULT_TYPE_OVERRIDES)
    def index(self, req):
        """Return a list of default types."""

        context = req.environ['cinder.context']
        try:
            context.authorize(policy.GET_ALL_POLICY)
        except exception.NotAuthorized:
            explanation = _("You are not authorized to perform this "
                            "operation.")
            raise exc.HTTPForbidden(explanation=explanation)

        default_types = db.project_default_volume_type_get(context)
        return self._view_builder.index(default_types)

    @wsgi.response(http_client.NO_CONTENT)
    @wsgi.Controller.api_version(mv.DEFAULT_TYPE_OVERRIDES)
    def delete(self, req, id):
        """Unset a default volume type for a project."""

        context = req.environ['cinder.context']

        project_id = id
        self._validate_project_and_authorize(context, project_id,
                                             policy.DELETE_POLICY)
        db.project_default_volume_type_unset(context, id)


def create_resource():
    """Create the wsgi resource for this controller."""
    return wsgi.Resource(DefaultTypesController())



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for V3 admin_actions API.

"""

import copy


from cinder.api.validation import parameter_types


reset = {
    'type': 'object',
    'properties': {
        'os-reset_status': {
            'type': 'object',
            'format': 'validate_volume_reset_body',
            'properties': {
                'status': {'type': ['string', 'null'],
                           'format': 'volume_status'},
                'attach_status': {'type': ['string', 'null'],
                                  'format': 'volume_attach_status'},
                'migration_status': {'type': ['string', 'null'],
                                     'format': 'volume_migration_status'},
            },
            'additionalProperties': False,
        },
    },
    'required': ['os-reset_status'],
    'additionalProperties': False,
}

force_detach = {
    'type': 'object',
    'properties': {
        'os-force_detach': {
            'type': 'object',
            'properties': {
                'connector': {'type': ['string', 'object', 'null']},
                'attachment_id': {'type': ['string', 'null']}
            },
            'additionalProperties': False,
        },
    },
    'required': ['os-force_detach'],
    'additionalProperties': False,
}


migrate_volume = {
    'type': 'object',
    'properties': {
        'os-migrate_volume': {
            'type': 'object',
            'properties': {
                'host': {'type': 'string', 'maxLength': 255},
                'force_host_copy': parameter_types.boolean,
                'lock_volume': parameter_types.boolean,
            },
            'required': ['host'],
            'additionalProperties': False,
        },
    },
    'required': ['os-migrate_volume'],
    'additionalProperties': False,
}


migrate_volume_v316 = {
    'type': 'object',
    'properties': {
        'os-migrate_volume': {
            'type': 'object',
            'properties': {
                'host': {'type': ['string', 'null'],
                         'maxLength': 255},
                'force_host_copy': parameter_types.boolean,
                'lock_volume': parameter_types.boolean,
                'cluster': parameter_types.name_allow_zero_min_length,
            },
            'additionalProperties': False,
        },
    },
    'required': ['os-migrate_volume'],
    'additionalProperties': False,
}


migrate_volume_completion = {
    'type': 'object',
    'properties': {
        'os-migrate_volume_completion': {
            'type': 'object',
            'properties': {
                'new_volume': parameter_types.uuid,
                'error': {'type': ['string', 'null', 'boolean']},
            },
            'required': ['new_volume'],
            'additionalProperties': False,
        },
    },
    'required': ['os-migrate_volume_completion'],
    'additionalProperties': False,
}


reset_status_backup = {
    'type': 'object',
    'properties': {
        'os-reset_status': {
            'type': 'object',
            'properties': {
                'status': {'type': 'string',
                           'format': 'backup_status'},
            },
            'required': ['status'],
            'additionalProperties': False,
        },
    },
    'required': ['os-reset_status'],
    'additionalProperties': False,
}

reset_status_snapshot = copy.deepcopy(reset_status_backup)
reset_status_snapshot['properties']['os-reset_status'][
    'properties']['status']['format'] = 'snapshot_status'



# Copyright 2015 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
Volume driver for IBM FlashSystem storage systems with iSCSI protocol.

Limitations:
1. Cinder driver only works when open_access_enabled=off.

"""

import random
import threading

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume import configuration as conf
from cinder.volume.drivers.ibm import flashsystem_common as fscommon
from cinder.volume.drivers.san import san

LOG = logging.getLogger(__name__)

flashsystem_iscsi_opts = [
    cfg.IntOpt('flashsystem_iscsi_portid',
               default=0,
               help='Default iSCSI Port ID of FlashSystem. '
                    '(Default port is 0.)')
]

CONF = cfg.CONF
CONF.register_opts(flashsystem_iscsi_opts, group=conf.SHARED_CONF_GROUP)


@interface.volumedriver
class FlashSystemISCSIDriver(fscommon.FlashSystemDriver):
    """IBM FlashSystem iSCSI volume driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.0.1 - Code clean up
        1.0.2 - Add lock into vdisk map/unmap, connection
                initialize/terminate
        1.0.3 - Initial driver for iSCSI
        1.0.4 - Split Flashsystem driver into common and FC
        1.0.5 - Report capability of volume multiattach
        1.0.6 - Fix bug #1469581, add I/T mapping check in
                terminate_connection
        1.0.7 - Fix bug #1505477, add host name check in
                _find_host_exhaustive for FC
        1.0.8 - Fix bug #1572743, multi-attach attribute
                should not be hardcoded, only in iSCSI
        1.0.9 - Fix bug #1570574, Cleanup host resource
                leaking, changes only in iSCSI
        1.0.10 - Fix bug #1585085, add host name check in
                 _find_host_exhaustive for iSCSI
        1.0.11 - Update driver to use ABC metaclasses
        1.0.12 - Update driver to support Manage/Unmanage
                 existing volume
    """

    VERSION = "1.0.12"

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "IBM_STORAGE_CI"

    def __init__(self, *args, **kwargs):
        super(FlashSystemISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(fscommon.flashsystem_opts)
        self.configuration.append_config_values(flashsystem_iscsi_opts)
        self.configuration.append_config_values(san.san_opts)

    def _check_vdisk_params(self, params):
        # Check that the requested protocol is enabled
        if not params['protocol'] in self._protocol:
            msg = (_("'%(prot)s' is invalid for "
                     "flashsystem_connection_protocol "
                     "in config file. valid value(s) are "
                     "%(enabled)s.")
                   % {'prot': params['protocol'],
                      'enabled': self._protocol})
            raise exception.InvalidInput(reason=msg)

        # Check if iscsi_ip is set when protocol is iSCSI
        if params['protocol'] == 'iSCSI' and params['iscsi_ip'] == 'None':
            msg = _("target_ip_address must be set in config file when "
                    "using protocol 'iSCSI'.")
            raise exception.InvalidInput(reason=msg)

    def _create_host(self, connector):
        """Create a new host on the storage system.

        We create a host and associate it with the given connection
        information.
        """

        LOG.debug('enter: _create_host: host %s.', connector['host'])

        rand_id = six.text_type(random.randint(0, 99999999)).zfill(8)
        host_name = '%s-%s' % (self._connector_to_hostname_prefix(connector),
                               rand_id)

        ports = []

        if 'iSCSI' == self._protocol and 'initiator' in connector:
            ports.append('-iscsiname %s' % connector['initiator'])

        self._driver_assert(ports,
                            (_('_create_host: No connector ports.')))
        port1 = ports.pop(0)
        arg_name, arg_val = port1.split()
        ssh_cmd = ['svctask', 'mkhost', '-force', arg_name, arg_val, '-name',
                   '"%s"' % host_name]
        out, err = self._ssh(ssh_cmd)
        self._assert_ssh_return('successfully created' in out,
                                '_create_host', ssh_cmd, out, err)

        for port in ports:
            arg_name, arg_val = port.split()
            ssh_cmd = ['svctask', 'addhostport', '-force',
                       arg_name, arg_val, host_name]
            out, err = self._ssh(ssh_cmd)
            self._assert_ssh_return(
                (not out.strip()),
                '_create_host', ssh_cmd, out, err)

        LOG.debug(
            'leave: _create_host: host %(host)s - %(host_name)s.',
            {'host': connector['host'], 'host_name': host_name})

        return host_name

    def _find_host_exhaustive(self, connector, hosts):
        LOG.debug('enter: _find_host_exhaustive hosts: %s.', hosts)
        hname = connector['host']
        hnames = [ihost[0:ihost.rfind('-')] for ihost in hosts]
        if hname in hnames:
            host = hosts[hnames.index(hname)]
            ssh_cmd = ['svcinfo', 'lshost', '-delim', '!', host]
            out, err = self._ssh(ssh_cmd)
            self._assert_ssh_return(
                out.strip(),
                '_find_host_exhaustive', ssh_cmd, out, err)
            for attr_line in out.split('\n'):
                attr_name, foo, attr_val = attr_line.partition('!')
                if (attr_name == 'iscsi_name' and
                        'initiator' in connector and
                        attr_val == connector['initiator']):
                    LOG.debug(
                        'leave: _find_host_exhaustive connector: %s.',
                        connector)
                    return host
        else:
            LOG.warning('Host %(host)s was not found on backend storage.',
                        {'host': hname})
        return None

    def _get_vdisk_map_properties(
            self, connector, lun_id, vdisk_name, vdisk_id, vdisk_params):
        """Get the map properties of vdisk."""

        LOG.debug(
            'enter: _get_vdisk_map_properties: vdisk '
            '%(vdisk_name)s.', {'vdisk_name': vdisk_name})

        preferred_node = '0'
        IO_group = '0'

        # Get preferred node and other nodes in I/O group
        preferred_node_entry = None
        io_group_nodes = []
        for k, node in self._storage_nodes.items():
            if vdisk_params['protocol'] != node['protocol']:
                continue
            if node['id'] == preferred_node:
                preferred_node_entry = node
            if node['IO_group'] == IO_group:
                io_group_nodes.append(node)

        if not io_group_nodes:
            msg = (_('No node found in I/O group %(gid)s for volume %(vol)s.')
                   % {'gid': IO_group, 'vol': vdisk_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not preferred_node_entry:
            # Get 1st node in I/O group
            preferred_node_entry = io_group_nodes[0]
            LOG.warning('_get_vdisk_map_properties: Did not find a '
                        'preferred node for vdisk %s.', vdisk_name)
        properties = {
            'target_discovered': False,
            'target_lun': lun_id,
            'volume_id': vdisk_id,
        }

        type_str = 'iscsi'
        if preferred_node_entry['ipv4']:
            ipaddr = preferred_node_entry['ipv4'][0]
        else:
            ipaddr = preferred_node_entry['ipv6'][0]
        iscsi_port = self.configuration.target_port
        properties['target_portal'] = '%s:%s' % (ipaddr, iscsi_port)
        properties['target_iqn'] = preferred_node_entry['iscsi_name']

        LOG.debug(
            'leave: _get_vdisk_map_properties: vdisk '
            '%(vdisk_name)s.', {'vdisk_name': vdisk_name})

        return {'driver_volume_type': type_str, 'data': properties}

    @utils.synchronized('flashsystem-init-conn', external=True)
    def initialize_connection(self, volume, connector):
        """Perform work so that an iSCSI connection can be made.

        To be able to create an iSCSI connection from a given host to a
        volume, we must:
        1. Translate the given iSCSI name to a host name
        2. Create new host on the storage system if it does not yet exist
        3. Map the volume to the host if it is not already done
        4. Return the connection information for relevant nodes (in the
        proper I/O group)

        """

        LOG.debug(
            'enter: initialize_connection: volume %(vol)s with '
            'connector %(conn)s.', {'vol': volume, 'conn': connector})

        vdisk_name = volume['name']
        vdisk_id = volume['id']
        vdisk_params = self._get_vdisk_params(volume['volume_type_id'])

        self._wait_vdisk_copy_completed(vdisk_name)

        self._driver_assert(
            self._is_vdisk_defined(vdisk_name),
            (_('vdisk %s is not defined.')
             % vdisk_name))

        lun_id = self._map_vdisk_to_host(vdisk_name, connector)

        properties = {}
        try:
            properties = self._get_vdisk_map_properties(
                connector, lun_id, vdisk_name, vdisk_id, vdisk_params)
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                self.terminate_connection(volume, connector)
                LOG.error('Failed to collect return properties for '
                          'volume %(vol)s and connector %(conn)s.',
                          {'vol': volume, 'conn': connector})

        LOG.debug(
            'leave: initialize_connection:\n volume: %(vol)s\n connector '
            '%(conn)s\n properties: %(prop)s.',
            {'vol': volume,
             'conn': connector,
             'prop': properties})

        return properties

    @utils.synchronized('flashsystem-term-conn', external=True)
    def terminate_connection(self, volume, connector, **kwargs):
        """Cleanup after connection has been terminated.

        When we clean up a terminated connection between a given connector
        and volume, we:
        1. Translate the given connector to a host name
        2. Remove the volume-to-host mapping if it exists
        3. Delete the host if it has no more mappings (hosts are created
        automatically by this driver when mappings are created)
        """
        LOG.debug(
            'enter: terminate_connection: volume %(vol)s with '
            'connector %(conn)s.',
            {'vol': volume, 'conn': connector})

        vdisk_name = volume['name']
        self._wait_vdisk_copy_completed(vdisk_name)
        host_name = self._unmap_vdisk_from_host(vdisk_name, connector)
        # checking if host_name none, if not then, check if the host has
        # any mappings, if not the host gets deleted.
        if host_name:
            if not self._get_hostvdisk_mappings(host_name):
                self._delete_host(host_name)

        LOG.debug(
            'leave: terminate_connection: volume %(vol)s with '
            'connector %(conn)s.', {'vol': volume, 'conn': connector})

        return {'driver_volume_type': 'iscsi'}

    def _get_iscsi_ip_addrs(self):
        """get ip address of iSCSI interface."""

        LOG.debug('enter: _get_iscsi_ip_addrs')

        cmd = ['svcinfo', 'lsportip']
        generator = self._port_conf_generator(cmd)
        header = next(generator, None)
        if not header:
            return

        for key in self._storage_nodes:
            if self._storage_nodes[key]['config_node'] == 'yes':
                node = self._storage_nodes[key]
                break

        if node is None:
            msg = _('No config node found.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        for port_data in generator:
            try:
                port_ipv4 = port_data['IP_address']
                port_ipv6 = port_data['IP_address_6']
                state = port_data['state']
                speed = port_data['speed']
            except KeyError:
                self._handle_keyerror('lsportip', header)
            if port_ipv4 == self.configuration.target_ip_address and (
                    port_data['id'] == (
                        six.text_type(
                            self.configuration.flashsystem_iscsi_portid))):
                if state not in ('configured', 'online'):
                    msg = (_('State of node is wrong. Current state is %s.')
                           % state)
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                if state in ('configured', 'online') and speed != 'NONE':
                    if port_ipv4:
                        node['ipv4'].append(port_ipv4)
                    if port_ipv6:
                        node['ipv6'].append(port_ipv6)
                    break
        if not (len(node['ipv4']) or len(node['ipv6'])):
            msg = _('No ip address found.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('leave: _get_iscsi_ip_addrs')

    def do_setup(self, ctxt):
        """Check that we have all configuration details from the storage."""

        LOG.debug('enter: do_setup')

        self._context = ctxt

        # Get data of configured node
        self._get_node_data()

        # Get the iSCSI IP addresses of the FlashSystem nodes
        self._get_iscsi_ip_addrs()

        for k, node in self._storage_nodes.items():
            if self.configuration.flashsystem_connection_protocol == 'iSCSI':
                if (len(node['ipv4']) or len(node['ipv6']) and
                        len(node['iscsi_name'])):
                    node['protocol'] = 'iSCSI'

        self._protocol = 'iSCSI'

        # Set for vdisk synchronization
        self._vdisk_copy_in_progress = set()
        self._vdisk_copy_lock = threading.Lock()
        self._check_lock_interval = 5

        LOG.debug('leave: do_setup')

    def _build_default_params(self):
        protocol = self.configuration.flashsystem_connection_protocol
        if protocol.lower() == 'iscsi':
            protocol = 'iSCSI'
        return {
            'protocol': protocol,
            'iscsi_ip': self.configuration.target_ip_address,
            'iscsi_port': self.configuration.target_port,
            'iscsi_ported': self.configuration.flashsystem_iscsi_portid,
        }

    def validate_connector(self, connector):
        """Check connector for enabled protocol."""
        valid = False
        if 'iSCSI' == self._protocol and 'initiator' in connector:
            valid = True
        if not valid:
            LOG.error('The connector does not contain the '
                      'required information: initiator is missing')
            raise exception.InvalidConnectorException(missing=(
                                                      'initiator'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import timeutils
import pytz

from cinder.db.sqlalchemy import models
from cinder import exception
from cinder import objects
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import objects as test_objects

fake_qos = {'consumer': 'front-end',
            'id': fake.OBJECT_ID,
            'name': 'qos_name',
            'specs': {'key1': 'val1', 'key2': 'val2'}}

fake_qos_no_id = fake_qos.copy()
del fake_qos_no_id['id']


class TestQos(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.get_by_id', return_value=fake_qos)
    def test_get_by_id(self, qos_get):
        qos_object = objects.QualityOfServiceSpecs.get_by_id(
            self.context, fake.OBJECT_ID)
        self._compare(self, fake_qos, qos_object)
        qos_get.assert_called_once_with(
            self.context, models.QualityOfServiceSpecs, fake.OBJECT_ID)

    @mock.patch('cinder.db.qos_specs_create',
                return_value={'name': 'qos_name', 'id': fake.OBJECT_ID})
    def test_create(self, qos_fake_create):
        qos_object = objects.QualityOfServiceSpecs(
            self.context, **fake_qos_no_id)
        qos_object.create()
        self._compare(self, fake_qos, qos_object)

        # Fail to create a second time
        self.assertRaises(exception.ObjectActionError, qos_object.create)

        self.assertEqual(1, len(qos_fake_create.mock_calls))

    @mock.patch('cinder.db.qos_specs_item_delete')
    @mock.patch('cinder.db.qos_specs_update')
    def test_save(self, qos_fake_update, qos_fake_delete):
        qos_dict = fake_qos.copy()
        qos_dict['specs']['key_to_remove1'] = 'val'
        qos_dict['specs']['key_to_remove2'] = 'val'
        qos_object = objects.QualityOfServiceSpecs._from_db_object(
            self.context, objects.QualityOfServiceSpecs(), qos_dict)

        qos_object.specs['key1'] = 'val1'
        qos_object.save()
        # No values have changed so no updates should be made
        self.assertFalse(qos_fake_update.called)

        qos_object.consumer = 'back-end'
        qos_object.specs['key1'] = 'val2'
        qos_object.specs['new_key'] = 'val3'

        del qos_object.specs['key_to_remove1']
        del qos_object.specs['key_to_remove2']
        qos_object.save()
        qos_fake_update.assert_called_once_with(
            self.context, fake.OBJECT_ID,
            {'specs': {'key1': 'val2', 'new_key': 'val3'},
             'consumer': 'back-end'})
        qos_fake_delete.assert_has_calls([
            mock.call(self.context, fake.OBJECT_ID, 'key_to_remove1'),
            mock.call(self.context, fake.OBJECT_ID, 'key_to_remove2')],
            any_order=True)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.objects.VolumeTypeList.get_all_types_for_qos',
                return_value=None)
    @mock.patch('cinder.db.sqlalchemy.api.qos_specs_delete')
    def test_destroy_no_vol_types(self, qos_fake_delete, fake_get_vol_types,
                                  utcnow_mock):
        qos_fake_delete.return_value = {
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        qos_object = objects.QualityOfServiceSpecs._from_db_object(
            self.context, objects.QualityOfServiceSpecs(), fake_qos)
        qos_object.destroy()

        qos_fake_delete.assert_called_once_with(mock.ANY, fake_qos['id'])
        self.assertTrue(qos_object.deleted)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         qos_object.deleted_at)

    @mock.patch('cinder.db.sqlalchemy.api.qos_specs_delete')
    @mock.patch('cinder.db.qos_specs_disassociate_all')
    @mock.patch('cinder.objects.VolumeTypeList.get_all_types_for_qos')
    def test_destroy_with_vol_types(self, fake_get_vol_types,
                                    qos_fake_disassociate, qos_fake_delete):
        qos_object = objects.QualityOfServiceSpecs._from_db_object(
            self.context, objects.QualityOfServiceSpecs(), fake_qos)
        fake_get_vol_types.return_value = objects.VolumeTypeList(
            objects=[objects.VolumeType(id=fake.VOLUME_TYPE_ID)])
        self.assertRaises(exception.QoSSpecsInUse, qos_object.destroy)

        qos_object.destroy(force=True)
        qos_fake_delete.assert_called_once_with(mock.ANY, fake_qos['id'])
        qos_fake_disassociate.assert_called_once_with(
            self.context, fake_qos['id'])

    @mock.patch('cinder.objects.VolumeTypeList.get_all_types_for_qos',
                return_value=None)
    @mock.patch('cinder.db.get_by_id', return_value=fake_qos)
    def test_get_volume_type(self, fake_get_by_id, fake_get_vol_types):
        qos_object = objects.QualityOfServiceSpecs.get_by_id(
            self.context, fake.OBJECT_ID)
        self.assertFalse(fake_get_vol_types.called)
        # Access lazy-loadable attribute
        qos_object.volume_types
        self.assertTrue(fake_get_vol_types.called)
